import dataclasses
import datetime
import typing
from functools import singledispatchmethod

from dict2dc.models.base import UNTOUCHED_TYPE
from dict2dc.models.requests import Params

D = typing.TypeVar("D")


class Dc2Query:
    """
    Converts a dataclass into a dictionary that can be used as query parameters for requests.
    """

    def as_query_params(self, obj: D | None) -> Params | None:
        """
        Converts a dataclass into a dictionary that can be used as query parameters for requests.
        It can only handle dataclasses that are not nested.
        It strips None and UNTOUCHED values.

        :param obj: the dataclass instance
        :return: the dictionary
        """
        if obj is None:
            return None
        obj_as_dict = dataclasses.asdict(obj)
        keys = set(obj_as_dict.keys())
        for key in keys:
            value = obj_as_dict[key]

            if value is None or isinstance(value, UNTOUCHED_TYPE):
                # Ignore None and untouched
                obj_as_dict.pop(key)
                continue
            else:
                updated_value = self._convert(value)

            obj_as_dict[key] = updated_value

        return obj_as_dict

    @singledispatchmethod
    def _convert(self, value: typing.Any) -> typing.Any:
        return value

    @_convert.register(bool)
    def _(self, value: bool) -> str:
        return str(value).lower()

    @_convert.register(datetime.date)
    @_convert.register(datetime.time)
    @_convert.register(datetime.datetime)
    def _(self, value: datetime.date | datetime.time | datetime.datetime) -> str:
        return value.isoformat()

    @_convert.register(set)
    def _(self, value: set) -> str | None:
        return ",".join(list(sorted(map(str, value))))

    @_convert.register(list)
    def _(self, value: list) -> str:
        return ",".join(map(str, value))



from collections.abc import Iterable
from typing import TypeAlias

# Use dict with str key as a pragmatic type that does not confuse the type checker
_ParamsMappingValueType: TypeAlias = str | bytes | int | float | Iterable[str | bytes | int | float] | None
Params: TypeAlias = dict[str, _ParamsMappingValueType]



import dataclasses
import datetime
import decimal
import json
import typing
import uuid
from functools import singledispatchmethod

UNSERIALIZABLE = object()


class Dc2Json:
    """
    Methods to transform an object to a JSON-serializable data structure.
    Heavily inspired from https://stackoverflow.com/a/51674892
    """

    @singledispatchmethod
    def as_serializable(self, obj: typing.Any):
        """
        Filter and transform a Python object to only include serializable object types

        In dictionaries, keys are converted to strings.

        :param obj: The object to serialize
        """

        # Dataclasses have no specific type, thus they must be handled here :(
        if dataclasses.is_dataclass(obj):
            return self.as_serializable(dataclasses.asdict(obj))

        # default handler, called for anything without a specific
        # type registration.
        return UNSERIALIZABLE

    @as_serializable.register(dict)
    def _handle_dict(self, d: dict):
        converted = ((str(k), self.as_serializable(v)) for k, v in d.items())
        return {k: v for k, v in converted if v is not UNSERIALIZABLE}

    @as_serializable.register(list)
    @as_serializable.register(tuple)
    def _handle_sequence(self, seq: list | tuple):
        converted = (self.as_serializable(v) for v in seq)
        return [v for v in converted if v is not UNSERIALIZABLE]

    @as_serializable.register(set)
    def _handle_set(self, seq: set):
        converted = sorted(self.as_serializable(v) for v in seq)
        return [v for v in converted if v is not UNSERIALIZABLE]

    @as_serializable.register(datetime.date)
    @as_serializable.register(datetime.time)
    @as_serializable.register(datetime.datetime)
    def _handle_date_and_time(self, value: datetime.date | datetime.time | datetime.datetime):
        return value.isoformat()

    @as_serializable.register(decimal.Decimal)
    def _handle_decimal(self, value: decimal.Decimal):
        # Return Decimals as strings
        return f"{value:.2f}"

    @as_serializable.register(uuid.UUID)
    def _handle_uuid(self, value: uuid.UUID):
        return str(value)

    @as_serializable.register(int)
    @as_serializable.register(float)
    @as_serializable.register(str)
    @as_serializable.register(bool)  # redundant, supported as int subclass
    @as_serializable.register(type(None))
    def _handle_default_scalar_types(self, value):
        return value


class DcJsonEncoder(json.JSONEncoder):
    """
    JSONEncoder that can handle Dataclasses, date, time, datetime, and our UNTOUCHED_TYPE.
    """

    dc2json = Dc2Json()

    def default(self, obj):
        return self.dc2json.as_serializable(obj)



import collections.abc
import dataclasses
import datetime
import functools
import logging
from types import NoneType

import types
import typing
from collections.abc import Collection, Callable
from inspect import isabstract

from dict2dc.models.base import UNTOUCHED_TYPE
from dict2dc.models.parsing import DictToDataclassConversionError, FieldResult
from dict2dc.utils.types import is_union

D = typing.TypeVar("D")
D2 = typing.TypeVar("D2")

LOGGER = logging.getLogger("dict2dc")

DEFAULT_CONVERSION: dict[type, Callable[[typing.Any], typing.Any]] = {
    datetime.datetime: datetime.datetime.fromisoformat,
    datetime.date: datetime.date.fromisoformat,
    datetime.time: datetime.time.fromisoformat,
}


class Dict2Dc:
    """
    Our class to parse dicts to dataclasses.
    """

    def __init__(
        self, special_conversions: dict[type, Callable[[typing.Any], typing.Any]] | None = None, replace: bool = False
    ) -> None:
        """
        :param special_conversions: Special conversions for some types.
        :param replace: Whether the default conversions should be replaced or not updated.
        """
        self._special_conversions: dict[type, Callable[[typing.Any], typing.Any]] = DEFAULT_CONVERSION.copy()
        if special_conversions is not None:
            if replace:
                self._special_conversions = special_conversions
            else:
                self._special_conversions.update(special_conversions)
        super().__init__()

    @staticmethod
    def ensure_dataclass(obj: typing.Any):
        """
        Checks if the given object is a dataclass (or an instance) and raises a TypeError otherwise.

        :param obj: the object to check
        :raise TypeError if the given object is not a dataclass
        """
        if not dataclasses.is_dataclass(obj):
            raise TypeError(f"{obj} is not a dataclass")

    def _handle_field(self, field_name: str, field_value: typing.Any, cls_: type) -> FieldResult:  # noqa: PLR0911
        """
        :param field_name: the name of the field
        :param field_value: the value to check
        :param cls_: the type it should match
        :return: whether the value matches the given type
        """
        self._ensure_type(cls_)

        # if the type has a generic, isinstance does not work; thus we need the type without the generic
        # example: list[str] → list
        cls_without_generics = cls_
        while isinstance(cls_without_generics, types.GenericAlias):
            cls_without_generics = typing.get_origin(cls_without_generics)

        if dataclasses.is_dataclass(cls_):
            parsed_value = self._handle_dataclass(field_value, cls_)
            if parsed_value is not None:
                return parsed_value

        union_result = self._handle_union(field_name, field_value, cls_)
        if union_result is not None:
            return union_result

        sequence_result = self._handle_collection(field_value, field_name, cls_)
        if sequence_result is not None:
            return sequence_result

        if typing.get_origin(cls_) is typing.Literal:
            is_valid_literal = field_value in typing.get_args(cls_)
            return FieldResult(
                value=field_value,
                type_matches=is_valid_literal,
                matching_score=1 if is_valid_literal else 0,
                error=None
                if is_valid_literal
                else DictToDataclassConversionError.from_parsing_error(
                    field_name=field_name, field_type=cls_, field_value=field_value
                ),
            )

        special_class_result = self._handle_special_classes(cls_, field_value)
        if special_class_result is not None:
            return special_class_result

        constructor_result = self._handle_constructor(cls_, cls_without_generics, field_value)
        if constructor_result is not None:
            return constructor_result

        type_matches = isinstance(field_value, cls_without_generics)

        return FieldResult(
            value=field_value,
            type_matches=type_matches,
            matching_score=1 if type_matches else 0,
            error=None
            if type_matches
            else DictToDataclassConversionError.from_parsing_error(field_name, cls_, field_value),
        )

    def _handle_collection(self, field_value: typing.Any, field_name: str, cls_: type) -> FieldResult | None:
        type_origin: type | None = typing.get_origin(cls_)
        # typing.Literal is not a class that can be used with issubclass - don't know why
        if type_origin is not None and type_origin is not typing.Literal and issubclass(type_origin, Collection):
            is_concrete_type = not isabstract(type_origin)

            if isinstance(field_value, type_origin if is_concrete_type else Collection) or (
                # Special handling for tuples and sets:
                # In JSON, there is no representation for these, thus we convert them from lists.
                isinstance(field_value, list) and type_origin in (tuple, set)
            ):
                typing_args = typing.get_args(cls_)
                item_type = typing_args[0]

                # We currently support tuples that have an arbitrary length, but each element has the same type,
                # e.g., tuple[str, ...]
                # (see https://docs.python.org/3/library/typing.html#annotating-tuples)
                if len(typing_args) == 1 or (
                    type_origin is tuple and len(typing_args) == 2 and typing_args[1] is Ellipsis  # noqa: PLR2004
                ):
                    results: list[FieldResult] = [
                        self._handle_field(field_name=f"{field_name}[{idx}]", cls_=item_type, field_value=item)
                        for idx, item in enumerate(field_value)
                    ]
                    # All items must have the correct type
                    if not all(result.type_matches for result in results):
                        return FieldResult(
                            value=field_value,
                            type_matches=False,
                            error=DictToDataclassConversionError(
                                ", ".join(str(result.error) for result in results if result.error is not None),
                            ),
                        )

                    matching_score = 1 + sum(result.matching_score for result in results)
                    if is_concrete_type:
                        # noinspection PyArgumentList
                        return FieldResult(
                            value=type_origin(result.value for result in results),
                            type_matches=True,
                            matching_score=matching_score,
                        )
                    else:
                        # for backwards-compatibility,
                        # we return a list for a generic Collection or Sequence instead of a tuple
                        return FieldResult(
                            value=[result.value for result in results], type_matches=True, matching_score=matching_score
                        )
        # not applicable
        return None

    def _handle_dataclass(self, d: dict | None, cls_: type[D]) -> FieldResult | None:
        if d is None:
            return FieldResult(
                type_matches=False,
                error=DictToDataclassConversionError(
                    f'Field in unexpectedly None and thus cannot be parsed as "{cls_}"'
                ),
            )

        d_normalized = {}  # the parts of the dictionary that belong to the dataclass
        matching_score = 0
        for field in dataclasses.fields(cls_):
            is_required_field = field.default == dataclasses.MISSING and field.default_factory == dataclasses.MISSING
            field_name = field.name
            field_exists_in_d = field_name in d

            if not field_exists_in_d:
                field_exists_in_d = self._handle_space_in_key(d, field_name)

            # Ensure that a required field exists, ...
            if is_required_field and not field_exists_in_d:
                return FieldResult(
                    type_matches=False,
                    error=DictToDataclassConversionError(f'Required field "{field_name}" is missing'),
                )
            initial_field_value = d.get(field_name, None)
            field_value = initial_field_value
            # pyre-ignore[6]: pyre does not recognize a type as Hashable
            field_type = self._resolve_type(cls_, field)

            if field_exists_in_d:
                result = self._handle_field(field_name, field_value, field_type)
                matching_score += result.matching_score

                if not result.type_matches:
                    return FieldResult(
                        value=field_value,
                        type_matches=False,
                        error=DictToDataclassConversionError.from_parsing_error(field_name, field_type, field_value),
                    )

                d_normalized[field_name] = result.value

        return FieldResult(value=cls_(**d_normalized), type_matches=True, matching_score=matching_score)

    def _handle_union(self, field_name: str, field_value: typing.Any, cls_: type) -> FieldResult | None:
        if is_union(cls_):
            union_types = typing.get_args(cls_)

            # Special case: We've got empty dicts instead of None in some cases
            empty_dict_might_be_none = any(t is types.NoneType for t in union_types) and all(
                t is not dict for t in union_types
            )

            matching_values: list[FieldResult] = []
            for t in union_types:
                result = self._handle_field(field_name, field_value, t)
                if result.type_matches:
                    matching_values.append(result)
            if len(matching_values) > 0:
                # We use the best-fitting union type
                return max(matching_values, key=lambda r: r.matching_score)

            if empty_dict_might_be_none and isinstance(field_value, dict) and field_value == {}:
                return FieldResult(value=None, type_matches=True)

            return FieldResult(
                value=field_value,
                type_matches=False,
                error=DictToDataclassConversionError.from_parsing_error(field_name, cls_, field_value),
            )
        return None

    @staticmethod
    def _ensure_type(cls_: typing.Any):
        """
        Ensures that the given class is a type that we can handle.
        :param cls_: the thing to check
        :raise TypeError if we cannot handle this type
        """
        if not (
            isinstance(cls_, type)
            or is_union(cls_)
            or typing.get_origin(cls_) is typing.Literal
            or isinstance(cls_, types.GenericAlias)
        ):
            raise TypeError(f"Class parameter {cls_!r} of type {type(cls_)} is not a valid type.")

    @staticmethod
    def _handle_constructor(cls_: type, cls_without_generics: type, value: typing.Any) -> FieldResult | None:
        if (
            cls_ == UNTOUCHED_TYPE
            or issubclass(cls_without_generics, collections.abc.Iterable)
            or cls_ == NoneType
            or not isinstance(value, str)
        ):
            return None
        try:
            value = cls_(value)
            return FieldResult(value=value, type_matches=True, matching_score=1)
        except (ValueError, TypeError):
            LOGGER.debug("Value %s cannot be parsed as %s", value, cls_)
        return None

    def _handle_special_classes(self, cls_: type, value: typing.Any) -> FieldResult | None:
        # Try to parse date/time values
        if cls_ in self._special_conversions and isinstance(value, str):
            try:
                value = self._special_conversions[cls_](value)
                return FieldResult(value=value, type_matches=True, matching_score=1)
            except ValueError:
                LOGGER.debug("Value %s cannot be parsed as %s", value, cls_)
        return None

    def from_dict_enforced(self, d: dict | None, cls: type[D]) -> D:
        """
        Wrapper around `from_dict` for better type checking by skipping the None in the return type.

        :param d: the dict
        :param cls: the class to create
        :return: the instance of the class
        """
        result = self.from_dict(d, cls, enforce=True)
        # This is never the case since an exception is thrown in from_dict when enforce is set to True
        if result is None:
            raise ValueError("Illegal state")

        return result

    def from_dict(self, d: dict | None, cls: type[D], enforce=False) -> D | None:
        """
        Creates the given dataclass `cls` from the given dict
        if it contains all the required fields with the correct type.

        :param d: the dict
        :param cls: the class to create
        :param enforce: whether the given dict must be converted to the class and produce a TypeError otherwise.
                        When false, None is returned
        :return: the instance of the class if all required fields have been found, None otherwise
        """
        self.ensure_dataclass(cls)

        return self.from_any(d, cls, enforce=enforce)

    def from_any(self, d: typing.Any, cls: type[D], enforce=False) -> D:
        """
        Tries to create the given class `cls` from the given data.
        This class might be a dataclass, a list of such or some plain types.

        :param d: The data to parse.
        :param cls: The resulting class.
        :param enforce: Whether the given data must be converted to the class and produce a TypeError otherwise.
                        When false, None is returned.
        :return: The instance of the class if all required fields have been found, None otherwise.
        """
        result = self._handle_field(".", d, cls)

        if not result.type_matches:
            if not enforce:
                LOGGER.debug(
                    "The given dict could not be parsed as %s due to: %s. Provided dict was: %s", cls, result.error, d
                )
                return None

            raise TypeError(
                f"The given dict could not be parsed as {cls} due to: {result.error}. Provided dict was: {d}"
            )

        return result.value

    @staticmethod
    def _handle_space_in_key(d: dict, field_name: str) -> bool:
        # Sometimes, the JSON has a space ins keys that are probably mapped to underscores in the dataclasses.
        # Override this method if you need another handling.
        if "_" in field_name:
            field_name_with_spaces = field_name.replace("_", " ")
            if field_name_with_spaces in d:
                d[field_name] = d[field_name_with_spaces]
                return True
        return False

    def _handle_field_and_ensure_type(self, field_name: str, field_type: type, field_value: typing.Any):
        field_value, type_matches = self._handle_field(field_name, field_value, field_type)

        if not type_matches:
            raise DictToDataclassConversionError.from_parsing_error(field_name, field_type, field_value)

        return field_value, type_matches

    @functools.cache
    def _resolve_type(self, cls_: type | str, field: dataclasses.Field) -> type:
        """
        Resolves the type of the given class field.
        Returns the type of field if it is already a type.
        Otherwise, looks up the type hint on the class and returns the type of the class field.

        Background:
        If we use from __future__ import annotations (PEP 563 Postponed Evaluation of Annotations),
        we can use field types in our models before the classes are defined.
        This leads to the circumstance that dataclasses.field.type returns the type as string instead of the actual
        class reference.

        This operation is expensive, thus we cache the result.

        :param cls_: the class
        :param field: the field
        :return: the resolved type
        """
        field_type = field.type
        if isinstance(field_type, type):
            return field_type

        resolved = typing.get_type_hints(cls_)
        field_type = resolved[field.name]
        return field_type

    def map_to(self, d: D2, cls: type[D]) -> D | None:
        """
        Converts an instance of one dataclass to another.

        :param d: the instance of a dataclass that should be mapped
        :param cls: the type of the dataclass to which should be mapped
        :return: the mapped dataclass instance
        """
        self.ensure_dataclass(d)
        dataclass_as_dict = dataclasses.asdict(d)
        return self.from_dict(dataclass_as_dict, cls, enforce=True)



# A sentinel object to detect if a field of a patch request class was touched or not.
# This way we can distinguish if it is the default value None or if the field is explicitly set to None.
# noinspection PyPep8Naming
class UNTOUCHED_TYPE:  # noqa: N801
    def __str__(self):
        return "<UNTOUCHED>"

    def __repr__(self):
        return "<UNTOUCHED>"

    def __bool__(self):
        return False


UNTOUCHED = UNTOUCHED_TYPE()



import dataclasses
import typing


class DictToDataclassConversionError(Exception):
    """
    An internal exception that might be thrown during the conversion of a dict to a dataclass in ~from_dict.
    """

    @classmethod
    def from_parsing_error(
        cls, field_name: str, field_type: type, field_value: typing.Any
    ) -> "DictToDataclassConversionError":
        return cls(f'Field "{field_name}" with type "{field_type}" cannot be parsed from "{field_value}"')


@dataclasses.dataclass
class FieldResult:
    """
    The (updated/parsed) value for the field.
    """

    value: typing.Any | None = None

    """
    Whether the value matches the field type.
    """
    type_matches: bool = False

    """
    How many sub fields have been parsed (without default values)
    """
    matching_score: int = 0

    """
    The error when the value does not match the field type.
    """
    error: DictToDataclassConversionError | None = None



from types import UnionType
from typing import TypeVar, Union, get_origin

T = TypeVar("T")


def is_union(t: type) -> bool:
    """
    Borrowed from https://stackoverflow.com/a/74546180

    :param t: the type to check
    :return: whether the type is a union type
    """
    origin = get_origin(t)
    return origin is Union or origin is UnionType
